"""
    Main Module of autosync Application
"""
import asyncio
import datetime
import sys
import time

import pandas as pd
from tabulate import tabulate

from autosync import const, lib, model, utils
# from autosync.const import appcfg, meraki_sdk
from autosync.model import golden_nets, meraki_nets


def completioninfo():
	"""
    Prints table of infomration after all tasks are complete
    Args:
    Returns:

    """
	orgcount = 0
	network_count = 0
	output_table = []
	for org in meraki_nets:
		orgcount = orgcount + 1
		network_count = network_count + int(len(meraki_nets[org].networks))
		output_table.append({
				'Orginization Name': meraki_nets[org].name,
				'Total Network'    : len(meraki_nets[org].networks),
				'Sync Runtime'     : meraki_nets[org].syncruntime,
				'Last Sync'        : meraki_nets[org].lastsync
		})
	print(f'Total Orgs Synced: {orgcount} '
	      f'Total Network Synced: {network_count}')
	table = pd.DataFrame.from_dict(output_table)
	print(tabulate(table, headers='keys', tablefmt='psql'))


def setup_app(cfg_file=None):
	const.appcfg = model.APPCONFIG(cfg_file)
	temp_sdk = lib.MerakiApi()
	const.meraki_sdk = temp_sdk.api


async def run(cfg_file=None):
	"""
    Main Module Start of Application
    Args:
    Returns:

    """
	setup_app(cfg_file)
	start_time = time.perf_counter()
	print(f'Start time: {start_time:0.5f}')
	golden_nets.update({
			const.appcfg.tag_golden:
				model.ORGDB(const.appcfg.tag_golden,
				            const.appcfg.tag_golden)
	})
	golden_nets[const.appcfg.tag_golden].networks.update(
			{const.appcfg.tag_golden: const.appcfg.tag_golden})
	if const.appcfg.use_cache:
		lib.load_cache(const.appcfg.tag_golden,True)
	if const.appcfg.all_orgs:
		orgs = utils.getOrginizationsAll()
		utils.set_orginization_by_id(orgs)
	elif len(const.appcfg.allow_org_list) != 0:
		orgdb_tasks = [
				utils.getOrginization(org)
				for org in const.appcfg.allow_org_list
		]
		await asyncio.gather(*orgdb_tasks)
	elif len(const.appcfg.allow_org_list_names) != 0:
		orgs = utils.getOrginizationsAll()
		utils.set_orginizations_by_name(orgs)
	else:
		print("Error No Orginizations List Exiting")
		sys.exit()
	org_threads = []
	validate_thread = []
	for org_id in meraki_nets:
		org_threads.append(utils.Orgsyncprocessor(org_id))
	for orgthread in org_threads:
		orgthread.start()
	for orgthread in org_threads:
		orgthread.join()
	
	for org in meraki_nets:
		validate_thread.append(utils.Validateorginization(org))
	for thread in validate_thread:
		thread.start()
	for thread in validate_thread:
		thread.join()
	
	if const.appcfg.use_cache:
		golden_nets[
			const.appcfg.tag_golden].lastsync = datetime.datetime.utcnow()
		lib.store_cache(const.appcfg.tag_golden,True)
	
	print(f'Total Job Runtime: '
	      f'{(time.perf_counter() - start_time):0.5f} secounds')
	completioninfo()


if __name__ == '__main__':
	config_file = 'testing/config.json'
	asyncio.run(run(config_file))
	print('Done')



from autosync import model,const
async def idFromName (listDicts, name):
	for ld in listDicts:
		if ld['name'] == name:
			return ld  # ld['groupPolicyId']
	return None


# returns object in list where "name" matches <name>
async def matchGidByName(listDicts, gpid):
	for ld in listDicts:
		if 'groupPolicyId' in ld and ld['groupPolicyId'] == gpid:
			return ld
	return None

def get_golden_group_policy():
	"""
    Gets Golden Group Policy Object
    Returns: REturns Golden Group Policy

    """
	golden_tag = const.appcfg.tag_golden
	return model.golden_nets[golden_tag].networks[golden_tag].dashboard[
		'networks'].NetworkGroupPolicies


def get_network_group_policy(org_id, net_id):
	"""
    Returns Netwok Group Policy Object
    Args:
        org_id:  Org ID
        net_id: Network ID

    Returns: Network Group Policy Object

    """
	return model.meraki_nets[org_id].networks[net_id].dashboard[
		'networks'].NetworkGroupPolicies


def find_rf_profile(profile_name: str, profiles):
	"""
	Finds A RF Profile Name in a list of RF Profile DICTs
	Args:
		profile_name: RF Profile Name
		profiles: RF Profiles

	Returns(bool): True if found False if not:

	"""
	for profile in profiles:
		if profile['name'] == profile_name:
			return True
	
	return False


def find_rf_profile_id_by_name(profile_name: str, profiles: list):
	"""

	Args:
		profile_name: Name of Profile to FIn
		profiles: List of RF Profiles

	Returns: RF Proilfe ID

	"""
	for profile in profiles:
		if profile['name'] == profile_name:
			return profile['id']
	return False


"""
Sync Function for product and task will call overides when needed dynamic
"""
import asyncio
import threading
from random import randrange
from meraki.exceptions import AsyncAPIError
from autosync import lib, const, model


def set_sync(org_id: str, net_id: str, product: str, is_golden: bool):
    """
    Args:
        org_id:
        net_id:
        product:
        is_golden:

    Returns:
    """
    golden_tag = const.appcfg.tag_golden
    if is_golden:
        action = [model.golden_nets[golden_tag].networks[golden_tag]]
    else:
        action = [
            model.meraki_nets[org_id].networks[net_id]]
    return action


async def sync(sdk: object, org_id: str, net_id: str, product: str,
               is_golden: bool):
    """
    Proforms a full sync of the _config object that is passed to the fuction
    using the meraki dashboard SDK
    Args:
        sdk(object): Meraki dashboard SDK Objeect
        org_id(str): Orginization ID
        net_id(str): Current Network ID
        product(str):  Meraki Product being synced I.E. switch, wirless
        is_golden(bool): Sync of Golden network
    Returns:
	   Nothing Updates the _Config Object that is passed
    """
    if const.appcfg.debug:
        print(f'Current Thread Name:{threading.currentThread().name} '
              f'Thread net_id:{threading.currentThread().native_id}')
    _maction = getattr(sdk, product)
    net = set_sync(org_id, net_id, product, is_golden)
    tasks = net[0].dashboard[product].settings()
    #o_func = getattr(net, f"functions[{product}]['get']")
    o_func = net[0].functions[product]['get']
    net_name = net[0].name

    for task in tasks:
        waiting = randrange(0, 2)
        await asyncio.sleep(waiting)
        if const.appcfg.debug:
            print(
                f'\t {lib.bc.OKGREEN}Network:{net_name}'
                f'{lib.bc.OKBLUE} Requesting Config Object P{product} - {task} '
                f'in Orginization {threading.currentThread().name} with '
                f'thread :{threading.currentThread().native_id} {lib.bc.Default}'
            )
        if task in o_func:
            try:
                #await eval(f'_config.Get_{setting}(sdk, net_id,_appcfg)')
                action = getattr(net[0].dashboard[product],f'Get_{task}')
                await action(sdk,net[0].net_id,net_name)
            except AsyncAPIError as apie:
                print(
                    f'\t {lib.bc.FAIL} Error Running Setting {task} '
                    f'{lib.bc.WARNING}Error Message: {str(apie)}{lib.bc.Default}'
                )
            except Exception as error:
                print(f'{lib.bc.FAIL}Network: {net_name} '
                      f'{lib.bc.WARNING}Error with Module: {str(error)} '
                      f'{lib.bc.Default}'
                      f'Running OVerride Function {task}')
        else:
            try:
                action = getattr(_maction, f'get{task}')
                value = await action(net[0].net_id)
                dashboard = [net[0].dashboard[product]]
                setattr(dashboard[0],task,
                        value)
            except AsyncAPIError as apie:
                print(
                    f'\t {lib.bc.FAIL} Error Running Setting {task} '
                    f'{lib.bc.WARNING}Error Message: {str(apie)}{lib.bc.Default}'
                )
            except Exception as error:
                print(f'{lib.bc.FAIL}Network: {net_name} '
                      f'{lib.bc.WARNING}Error with Module: {str(error)}'
                      f'{lib.bc.Default}')



"""
Network Product Setup for all configuration items to be synced
"""
import copy

from autosync import lib, const
from autosync.mnetutils.resync import re_sync


class networks:
    """
    Dashboard settings to sync,
    """
    def __init__(self):
        self.Network = None
        self.NetworkAlertsSettings = None
        self.NetworkGroupPolicies = None
        self.NetworkTrafficAnalysis = None
        self.NetworkSyslogServers = None
        self.NetworkSnmp = None
        self.NetworkWebhooksHttpServers = None
        self.NetworkName = None

    def settings(self):
        """
        Returns:
            Settings to cycle through when updating configurate
        """
        skip = ['re_sync', 'NetworkName']  # Settings to skip
        items = [i for i in self.__dict__.keys() if i not in skip]
        return items

    async def Update_Network(self, sdk: object, org_id: str, net_id: str,
                             net_name):
        """
        Updates Network settings in the meraki dashboard
        Args:
            sdk: Meraki SDK Object
            org_id: Current Org ID
            net_id: Curremt Metwork ID
            net_name: Network Name
        Returns:
        """
        golden = lib.get_golden('networks')
        # "No Seetins to Update here from master"
        if self.Network['timeZone'] == golden.Network['timeZone']:
            lib.print_matched(net_name, "time Zone", "Network")
        else:
            if const.appcfg.write:
                lib.print_update(net_name, "Time Zone", "Network")
                await sdk.networks.updateNetwork(
                    net_id, timeZone=golden.Network['timeZone'])
                await re_sync(sdk, org_id, net_id, 'Network', 'networks')

    async def Update_NetworkSyslogServers(self, sdk: object, org_id: str,
                                          net_id: str, net_name):
        """
        Compaires Network Syslog Servers Configuration to Master and Updates
        Configuration if anything changes on target network
        Args:
            sdk: Meraki SDK Object
            org_id: Current Org ID
            net_id: Curremt Metwork ID
            net_name: Network Name
        Returns:
            Nothing Updates meraki Dashboard
        """
        golden = lib.get_golden('networks')
        if not await lib.compare(golden.NetworkSyslogServers['servers'],
                                 self.NetworkSyslogServers['servers']):

            # Fixed the Kinda Works by using Hashing to compair
            if const.appcfg.write:
                lib.print_update(net_name, "Syslog Settiungs", "Network")
                await sdk.networks.updateNetworkSyslogServers(
                    net_id, **{'servers': []})
                await sdk.networks.updateNetworkSyslogServers(
                    net_id, **golden.NetworkSyslogServers)
                await re_sync(sdk, org_id, net_id, 'NetworkSyslogServers',
                              'networks')
        else:
            lib.print_matched(net_name, "SysLog Settings", "Network")

    async def Update_NetworkAlertsSettings(self, sdk: object, org_id: str,
                                           net_id: str, net_name):
        """
            Compaires Network Alert Configuration to Master and Updates
            Configuration if anything changes on target network
            Args:
	            sdk: Meraki SDK Object
	            org_id: Current Org ID
	            net_id: Curremt Metwork ID
	            net_name: Network Name
            Returns:
                Nothing Updates meraki Dashbioard
            """
        golden = lib.get_golden('networks')
        golden_network_alert_settings = sorted(golden.NetworkAlertsSettings['alerts'], key = lambda type: type['type'])
        network_alert_settings = sorted(self.NetworkAlertsSettings['alerts'], key = lambda type: type['type'])
        update = False
        if len(golden_network_alert_settings) == len(network_alert_settings):
            for i in range(0, len(golden_network_alert_settings)):
                if not await lib.compare(golden_network_alert_settings[i],network_alert_settings[i]):
                    update = True
        if update:
            if const.appcfg.write:
                    await sdk.networks.updateNetworkAlertsSettings(
                            net_id, **golden.NetworkAlertsSettings)
                    lib.print_update(net_name, "Alert Settings",
                                     "Network")
                    await re_sync(sdk, org_id, net_id,
                                  'NetworkAlertsSettings',
                                  'networks')
        else:
            lib.print_matched(net_name, "Alert Settings", "Network")

    async def Update_NetworkWebhooksHttpServers(self, sdk: object, org_id: str,
                                                net_id: str, net_name):
        """
            Compaires Network WebHook Configuration to Master and Updates
            Configuration if anything changes on target network
            Args:
	            sdk: Meraki SDK Object
	            org_id: Current Org ID
	            net_id: Curremt Metwork ID
	            net_name: Network Name
            Returns:
                Nothing Updates meraki Dashbioard
            """

        # Webhooks
        golden = lib.get_golden('networks')
        if not await lib.compare(golden.NetworkWebhooksHttpServers,
                                 self.NetworkWebhooksHttpServers):
            curr_list = []
            for cwh in self.NetworkWebhooksHttpServers:
                curr_list.append(cwh['name'])
            for mwh in golden.NetworkWebhooksHttpServers:
                if not mwh['name'] in curr_list:
                    if const.appcfg.write:
                        lib.print_update(net_name, "WebHooks", "Network")
                        mwh_tmp = copy.deepcopy(mwh)
                        mwh_tmp.pop('networkId')
                        await sdk.networks.createNetworkWebhooksHttpServer(
                            net_id, **mwh_tmp)
            await re_sync(sdk, org_id, net_id, 'NetworkWebhooksHttpServer',
                          'networks')
        else:
            lib.print_matched(net_name, "WebHooks", "Network")

    # noinspection PyBroadException
    async def Update_NetworkGroupPolicies(self, sdk: object, org_id: str,
                                          net_id: str, net_name):
        """
        Compaires Network Group Policy Configuration to Master and Updates
        Configuration if anything changes on target network
        Args:
	            sdk: Meraki SDK Object
	            org_id: Current Org ID
	            net_id: Curremt Metwork ID
	            net_name: Network Name
        Returns:
            Nothing Updates meraki Dashbioard
         """
        # Group Policies
        # TODO Move this to Batch Action Could be faster
        golden = lib.get_golden('networks')
        resync = False
        for golden_gp in golden.NetworkGroupPolicies:
            tempGP = copy.deepcopy(golden_gp)
            tempGP.pop('groupPolicyId')
            if const.appcfg.write:
                local_gp = await lib.idFromName(self.NetworkGroupPolicies,
                                                tempGP['name'])
                if local_gp is None:
                    print(f'\t\t{lib.bc.OKBLUE}Creating GP Policy named '
                          f'{tempGP["name"]}{lib.bc.ENDC}')
                    try:
                        lib.print_update(net_name, "Group Policies",
                                         "Network")
                        await sdk.networks.createNetworkGroupPolicy(
                            net_id, **tempGP)
                        resync = True
                    except Exception as error:
                        print(
                            f'{lib.bc.FAIL}ERROR: Cannot create GP '
                            f'policy named {tempGP["name"]} Error: {str(error)}'
                        )

                else:
                    local_gpid = local_gp['groupPolicyId']
                    tempGP['groupPolicyId'] = local_gpid
                    if not await lib.soft_compare(
                            tempGP, await lib.idFromName(
                                golden.NetworkGroupPolicies, tempGP['name'])):
                        lib.print_update(net_name, "Group Policies",
                                         "Network")
                        resync = True
                        await sdk.networks.updateNetworkGroupPolicy(
                            net_id, **tempGP)

        if const.appcfg.write:
            for localgp in self.NetworkGroupPolicies:
                golden_gp = await lib.idFromName(golden.NetworkGroupPolicies,
                                                 localgp['name'])
                if golden_gp is None:
                    print(
                        f'{lib.bc.WARNING}{net_name} - '
                        f'Group Piolicy \t\t{lib.bc.OKBLUE}Removing GP Policy named '
                        f'{localgp["name"]}{lib.bc.ENDC} not found in Master')
                    resync = True
                    await sdk.networks.deleteNetworkGroupPolicy(net_id, groupPolicyId=str(localgp['groupPolicyId']))

        if resync:
            await re_sync(sdk, org_id, net_id, 'NetworkGroupPolicies',
                          'networks')



import asyncio
import click

import model
from autosync.main import run
from autosync import const
from autosync.model import APPCONFIG
@click.group()
def cli():
    """ Tasks to start sync auto Sycn of Meraki Networks """
    pass


@click.command(help='Starts Meraki Dashboard Autosync')
@click.option('-a', '--allOrgs')
@click.option('-c', '--useCache')
@click.option('-d','--debug')
@click.option('-f','--configfile')
@click.option('-k', '--merakiApiKey')
@click.option('-m', '--goldenTag')
@click.option('-o', '--autoSyncOrgs')
@click.option('-s', '--suppressLogging')
@click.option('-t', '--targetTag')
@click.option('-T', '--cacheTimeOut')
@click.option('-w', '--write')
@click.option('--tagOverRide')
def start(suppresslogging=None, merakiapikey=None, write=None, allorgs=None, autosyncorgs=None,
          usecache=None, cachetimeout=None, goldentag=None, targettag=None, tagoverride=None,
          logginglevel=None,configfile=None,debug=None):
    if configfile is not None:
        asyncio.run(run(configfile))

    else:
        cfg = {'suppress_logging'        : suppresslogging,
               'MERAKI_DASHBOARD_API_KEY': merakiapikey,
               'write'                   : write,
               'whitelist'               : autosyncorgs,
               'tag_golden'              : goldentag,
               'tag_target'              : targettag,
               'all_orgs'                : allorgs,
               'use_cache'               : usecache,
               'cache_timeout'           : cachetimeout,
               'tag_override'            : tagoverride,
               'logging_level'           : logginglevel,
               'debug'                   : debug}
        
        asyncio.run(run(cfg))
        





cli.add_command(start)



"""
Meraki Orginization Data Class
"""
class ORGDB:
    """
    Meraki Orginization Data Class
    """
    def __init__(self, org_id, org_name):
        self.org_id = org_id
        self.name = org_name
        self.syncruntime = None
        self.lastsync = None
        self.cached = False
        self.networks = {}
        self.change_log = []

    def __dir__(self):
        return self.__dict__.keys()



import os

import click

cmd_folder = os.path.join(os.path.dirname(__file__), 'commands')
cmd_prefix = 'cmd_'

class CLI(click.MultiCommand):
    def list_commands(self, ctx):
        """
        Obtain a list of all available commands.

        :param ctx: Click context
        :return: List of sorted commands
        """
        commands = []

        for filename in os.listdir(cmd_folder):
            if filename.endswith('.py') and filename.startswith(cmd_prefix):
                commands.append(filename[4:-3])

        commands.sort()

        return commands

    def get_command(self, ctx, name):
        """
        Get a specific command by looking up the module.

        :param ctx: Click context
        :param name: Command name
        :return: Module's cli function
        """
        #try:
        ns = {}

        filename = os.path.join(cmd_folder, cmd_prefix + name + '.py')

        with open(filename) as f:
            code = compile(f.read(), filename, 'exec')
            eval(code, ns, ns)

        return ns['cli']
        #except Exception as e:
        #    print(f'Command Not Found please use autosync --help {str(e)}' )


@click.command(cls=CLI)
def cli():
    """ Commands to help manage your project. """
    pass



__title__ = 'Meraki_Auto_Sync'
__description__ = 'Auto Sync Meraki Network Dashboard cross Orginization based on network tags for master and target'
__url__ = 'http://devnet.cisco.com'
__download_url__ = 'https://devnet.cisco.com'
__author__ = 'Josh Lipton'
__author_email__ = 'jolipton@cisco.com'
__copyright__ = "Copyright (c) 2021 Cisco Systems, Inc."
__license__ = "MIT"
